/*-------------------------------------------------------------------*/
/* CBDA.C     Juha Kivijrvi  1999                                   */
/*                                                                   */
/* Deterministic annealing                                           */
/*                                                                   */
/*       Changelog:                                                  */
/*                                                                   */
/* 0.05  add only one vector at each iteration                       */
/* 0.04  fixed FuzzyMSE + other stuff                                */
/*                                                                   */
/*-------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <values.h>

#if defined(__MSDOS__)
#include <conio.h>
#endif

#define FACTFILE "cbda.fac"

#include "memctrl.h"
#include "parametr.c"
#include "random.h"
#include "file.h"
#include "cb.h"
#include "owntypes.h"
#include "solution.h"

#define  ITERATIONEPSILON  0.00000001
#define  COMPARISONEPSILON 0.0001

/*#define  ITERATIONEPSILON  0.01
#define  COMPARISONEPSILON 0.1*/

typedef struct { clock_t latestclock;
                 int     carry;       } CLOCKINFO;

/*--------------------------------------------------------------------*/

#define  ProgName       "CBDA"
#define  VersionNumber  "version 0.05"
#define  LastUpdated    "16.11.99"        /*  (by   Juha Kivijrvi) */



/*============================  P R I N T  ===============================*/

void PrintInfo(void) 
{
printf("%s     %s                (last modified %s)\n",ProgName,
                                             VersionNumber,LastUpdated);
printf("This program creates a codebook using the deterministic annealing.\n");
printf("       Usage: CBDA [options] TSfile outputfile \n\n");
printf("       options:\n");
PrintOptions();
printf("\n");
printf("%s     %s                                Juha Kivijrvi  1999\n",ProgName,VersionNumber);
}


/*-------------------------------------------------------------------*/


void PrintElapsedTime(CLOCKINFO* ci)
{
llong time_elapsed;
long h,m,s;

time_elapsed = ( (llong) clock() + ci->carry * (llong) UINT_MAX ) /
                  CLOCKS_PER_SEC;

h = time_elapsed / 3600;
m = (time_elapsed % 3600) / 60;
s = time_elapsed % 60;

if (Value(ShowProgress)>1)
    printf("Time elapsed: ");
printf("%ld:%02ld:%02ld\n",h, m, s);
}        


/*-------------------------------------------------------------------*/


void UpdateClock(CLOCKINFO* ci)
{
clock_t cl;

cl = clock();
if ( cl < ci->latestclock ) ci->carry++;
ci->latestclock = cl;
}


/*===========================  M I S C ==============================*/


void ModifyCodebookFileHeader(CODEBOOK *CB)
{
char newmethod[MaxGenMethodLength];

ClearGenerationMethod(CB);

sprintf(newmethod, "DA");

AddGenerationMethod(CB, newmethod);
}


/*================  V E C T O R S  ==================================*/


double DoubleVectorDistance(VECTORTYPE v1, double* v2, int Vsize)
{
/* Squared Euclidean distance */

double dist = 0;
int i;

for (i=0; i<Vsize; i++)
    dist += (v1[i] - v2[i]) * (v1[i] - v2[i]);

return dist;
}


/*--------------------------------------------------------------------*/


YESNO EqualDoubleVectors(double* v1, double* v2, int Vsize, double Epsilon)
{
int i;

for (i=0; i<Vsize; i++)
    if ( fabs( v1[i] - v2[i] ) > Epsilon )
        return NO;

return YES;
}


/*--------------------------------------------------------------------*/


void CopyDoubleVector(double* to, double* from, int Vsize)
{
int i;

for (i=0; i<Vsize; i++)
    to[i] = from[i];
}


/*--------------------------------------------------------------------*/


void TSDoubleCentroid(TRAININGSET* TS, double* v)
{
int i, j;

for (i=0; i<VectorSize(TS); i++)
    v[i] = 0.0;

for (i=0; i<BookSize(TS); i++)
    for (j=0; j<VectorSize(TS); j++)
        v[j] += VectorScalar(TS, i, j) * VectorFreq(TS, i);

for (i=0; i<VectorSize(TS); i++)
    v[i] /= TotalFreq(TS);
}


/*================  O P E R A T I O N  ==============================*/


void DuplicateRandomVector(TRAININGSET* TS, double** dCB,
			   double* Probs, int CBsize)
{
int i, j;

/* Perturbation is done by adding a random number in (-0.5, 0.5)
   to each (duplicate) vector component. No real reason for the choice. */

i = irand(0, CBsize-1);

for (j=0; j<VectorSize(TS); j++)
    dCB[CBsize][j] = dCB[i][j] + drand() - 0.5;
Probs[i] = Probs[CBsize] = Probs[i] / 2.0;
}


/*--------------------------------------------------------------------*/


double FuzzyMSE(TRAININGSET* TS, double** dCB, double** condp, int CBsize)
{
int m, n;
double fMSE = 0.0;

for (m=0; m<CBsize; m++)
    for (n=0; n<BookSize(TS); n++)
	{
	fMSE += condp[m][n] * DoubleVectorDistance(Vector(TS,n), dCB[m],
			      VectorSize(TS)) * VectorFreq(TS,n);
	}

return fMSE / (VectorSize(TS) * TotalFreq(TS));
}


/*--------------------------------------------------------------------*/


void OptimizeCodebook(TRAININGSET* TS, double** dCB, double** newdCB,
		      double* Probs, int CBsize, double Beta)
{
int i, m, n;
double **term, *termsum, **condp, *condpsum;
double mindist;
double *dists;

term = (double**) allocate(sizeof(double*) * CBsize);
condp = (double**) allocate(sizeof(double*) * CBsize);
for (i=0; i<CBsize; i++)
    {
    term[i] = (double*) allocate(sizeof(double) * BookSize(TS));
    condp[i] = (double*) allocate(sizeof(double) * BookSize(TS));
    }
termsum = (double*) allocate(sizeof(double) * BookSize(TS));
condpsum = (double*) allocate(sizeof(double) * CBsize);
dists = (double*) allocate(sizeof(double) * CBsize);

for (n=0; n<BookSize(TS); n++)
    {
    termsum[n] = 0;
    mindist = MAXDOUBLE;
    for (m=0; m<CBsize; m++)
	{
        dists[m] = DoubleVectorDistance(Vector(TS,n), dCB[m], VectorSize(TS));
	if (dists[m] < mindist) mindist = dists[m];
	}
    for (m=0; m<CBsize; m++)
	{
	/* mindist * Beta is added to prohibit underflows. It shouldn't
	   have an effect on the result. */
/* 	term[m][n] = Probs[m] * exp((mindist-dists[m]) * Beta); */
 	term[m][n] = Probs[m] * exp(-dists[m] * Beta);
	termsum[n] += term[m][n];
	}
    }
for (m=0; m<CBsize; m++)
    {
    condpsum[m] = 0;
    for (n=0; n<BookSize(TS); n++)
	{
	condp[m][n] = term[m][n] / termsum[n];
	condpsum[m] += condp[m][n] * VectorFreq(TS, n);
	}
    }
for (m=0; m<CBsize; m++)
    for (i=0; i<VectorSize(TS); i++)
        {
        newdCB[m][i] = 0;
        for (n=0; n<BookSize(TS); n++)
	    newdCB[m][i] += VectorScalar(TS,n,i) * VectorFreq(TS, n) *
	                    condp[m][n];
        newdCB[m][i] /= condpsum[m];
        }
for (m=0; m<CBsize; m++)
    Probs[m] = condpsum[m] / TotalFreq(TS);

if ( Value(ShowProgress) > 4 )
    printf("fuzzy MSE = %.12f\n", FuzzyMSE(TS, newdCB, condp, CBsize));

deallocate(termsum);
deallocate(condpsum);
for (i=0; i<CBsize; i++)
    {
    deallocate(term[i]);
    deallocate(condp[i]);
    }
deallocate(term);
deallocate(condp);
deallocate(dists);
}


/*--------------------------------------------------------------------*/


YESNO EqualDoubleCodebooks(TRAININGSET* TS, double** cb1, double** cb2,
			   int CBsize)
{
int i;

for (i=0; i<CBsize; i++)
    if ( ! ( EqualDoubleVectors( cb1[i], cb2[i], VectorSize(TS),
			         ITERATIONEPSILON) ) )
        return NO;

return YES;
}


/*--------------------------------------------------------------------*/


void IterateCodebook(TRAININGSET* TS, double** dCB, double* Probs,
		     int *CBsize, double Beta)
{
YESNO goon = YES;
double** newdCB;
int i, j;
int newCBsize;

newdCB = (double**) allocate(sizeof(double*) * 2 * (*CBsize));
for (i=0; i<2*(*CBsize); i++)
    newdCB[i] = (double*) allocate(sizeof(double) * VectorSize(TS));

if (*CBsize < Value(CodebookSize))
    {
    DuplicateRandomVector(TS, dCB, Probs, *CBsize);
    newCBsize = *CBsize + 1;
    }
else
    newCBsize = *CBsize;

while (goon)
    {
    goon = NO;
    OptimizeCodebook(TS, dCB, newdCB, Probs, newCBsize, Beta);
/*    printf("Probs: ");
    for(i=0; i<newCBsize; i++) printf("%f ", Probs[i]); printf("\n"); */
    if (!(EqualDoubleCodebooks(TS, dCB, newdCB, newCBsize))) goon = YES;
    for (i=0; i<newCBsize; i++)
        for (j=0; j<VectorSize(TS); j++)
	    dCB[i][j] = newdCB[i][j];
    }

for (i=newCBsize-1; i>=0; i--)
    for (j=0; j<i; j++)
	if ( EqualDoubleVectors(dCB[i], dCB[j], VectorSize(TS),
				COMPARISONEPSILON) )
	    {
	    if (i < newCBsize-1)
            CopyDoubleVector(dCB[i], dCB[newCBsize-1], VectorSize(TS));
	    newCBsize--;
            break;
	    }

for (i=0; i<2*(*CBsize); i++)
    deallocate(newdCB[i]);

deallocate(newdCB);

*CBsize = newCBsize;
}


/*--------------------------------------------------------------------*/


void IncreaseBeta(double *Beta)
{
(*Beta) += (*Beta) * Value(BetaIncrement) / 10000.0;
}


/*--------------------------------------------------------------------*/


void DiscretizeCodebook(TRAININGSET* TS, CODEBOOK* CB, double** dCB)
{
int i, j;

for (i=0; i<Value(CodebookSize); i++)
    for (j=0; j<VectorSize(TS); j++)
        VectorScalar(CB, i, j) = (VECTORELEMENT) dCB[i][j];

}


/*=====================  I N I T I A L I Z E  ==========================*/


void InitializeCodebook(TRAININGSET* TS, double*** dCB,
			double** Probs, int* CBsize, int *newCBsize)
{
int i;

*dCB = allocate( sizeof(double*) * 2 * Value(CodebookSize) );
*Probs = allocate( sizeof(double) * 2 * Value(CodebookSize) );

for (i=0; i<2*Value(CodebookSize); i++)
    (*dCB)[i] = allocate( sizeof(double) * VectorSize(TS) );

TSDoubleCentroid(TS, (*dCB)[0]);

*CBsize = 1;
*newCBsize = 0;

(*Probs)[0] = 1.0;
}


/*=====================  S H U T D O W N  ============================*/


void ShutDown(TRAININGSET* TS, CODEBOOK* CB, double** dCB, double* Probs)
              
{
int i;

FreeCodebook(TS);
FreeCodebook(CB);

for (i=0; i<2*Value(CodebookSize); i++)
    deallocate(dCB[i]);

deallocate(dCB);
deallocate(Probs);
}


/*===========================  M A I N  ==============================*/



int main(int argc, char* argv[])

{
char        TSFileName[80], OutFileName[80];

TRAININGSET TS;
CODEBOOK    CB;
double**    dCB;
int         CBsize, oldCBsize;
double*     Probs;
double      Beta;
CLOCKINFO   ci = {0,0};
int         i, j;

ParameterInfo paraminfo[2] =  { { TSFileName,  FormatNameTS, 0, INFILE  },
                                { OutFileName, FormatNameCB, 0, OUTFILE } };

ParseParameters( argc, argv, 2, paraminfo);

Beta = 1.0 / Value(InitBetaDivisor);

ReadTrainingSet(TSFileName, &TS); 
/* UnrollTrainingSet(&TS); */
CreateNewCodebook(&CB, Value(CodebookSize), &TS);

initrandom(Value(RandomSeed));

if (ExistFile(OutFileName) && Value(OverWrite)==NO)
    {
    printf("ERROR: can't write output file\n");
    exit(-1);
    }

InitializeCodebook(&TS, &dCB, &Probs, &CBsize, &oldCBsize);

while (Beta < 1.0 / Value(MaxBetaDivisor) )
    {
    UpdateClock(&ci);
    IterateCodebook(&TS, dCB, Probs, &CBsize, Beta);
    if ( CBsize == oldCBsize ) IncreaseBeta(&Beta);
    if (Value(ShowProgress) > 0)
        printf("Beta = %.6f  CBsize = %d\n", Beta, CBsize);
    if (Value(ShowProgress) == 2 || Value(ShowProgress) == 4
         || Value(ShowProgress) == 6)
	{
	printf("codebook:\n");
	for (i=0; i<CBsize; i++)
	    {
	    for (j=0;j <VectorSize(&TS); j++)
	        printf("%.2f ",dCB[i][j]);
	    printf("\n");
	    }
	printf("-----------------\n");
	}
    if (CBsize > Value(CodebookSize))
	{
	printf("Desired codebook size skipped. Try reducing Beta increment.\n");
        break;
	}
    else if (CBsize < oldCBsize)
	{
	printf("Codebook size has diminished. Quitting.\n");
        break;
	}
    else
	{
	DiscretizeCodebook(&TS, &CB, dCB);
	if ( Value(ShowProgress) > 2 )
	    printf("MSE = %.6f\n", AverageErrorCB(&TS, &CB, MSE));
	}
    oldCBsize = CBsize;
    }

/* VectorFreq:t pitisi laittaa kuntoon! */
ModifyCodebookFileHeader(&CB);

WriteCodebook(OutFileName, &CB, Value(OverWrite));
if (Value(ShowProgress) <= 2)
    printf("MSE = %.6f\n", AverageErrorCB(&TS, &CB, MSE));

ShutDown(&TS, &CB, dCB, Probs);

if (Value(ShowProgress)>0) PrintElapsedTime(&ci);
else printf("\n");
checkmemory();
return(0);
}
