/*-------------------------------------------------------------------*/
/* CBDAINT.C     Juha Kivijrvi  1999                                */
/*                                                                   */
/* Deterministic annealing - integer version                         */
/*                                                                   */
/*       Changelog:                                                  */
/*                                                                   */
/* 0.04  created from cbda.c 0.04 by changing doubles to ints        */
/*                                                                   */
/*-------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <values.h>

#if defined(__MSDOS__)
#include <conio.h>
#endif

#define FACTFILE "cbda.fac"

#include "memctrl.h"
#include "parametr.c"
#include "random.h"
#include "file.h"
#include "cb.h"
#include "owntypes.h"
#include "solution.h"

#define  THRESHOLD  20           /* this is a bit problematic */

/*--------------------------------------------------------------------*/

#define  ProgName       "CBDAINT"
#define  VersionNumber  "version 0.04"
#define  LastUpdated    "10.11.99"        /*  (by   Juha Kivijrvi) */



/*============================  P R I N T  ===============================*/

void PrintInfo(void) 
{
printf("%s     %s                (last modified %s)\n",ProgName,
                                             VersionNumber,LastUpdated);
printf("This program creates a codebook using the deterministic annealing.\n");
printf("       Usage: CBDAINT [options] TSfile outputfile \n\n");
printf("       options:\n");
PrintOptions();
printf("\n");
printf("%s     %s                                Juha Kivijrvi  1999\n",ProgName,VersionNumber);
}


/*-------------------------------------------------------------------*/


void OutputTime(time_t StartingTime)
{
time_t endtime;
long time_elapsed;
long processortime;
long h,m,s;


endtime=time(NULL);
time_elapsed=(long) difftime(endtime,StartingTime);
processortime=(long)(clock()/CLOCKS_PER_SEC);

h=time_elapsed/3600;
m=(time_elapsed % 3600)/60;
s=time_elapsed%60;

if (Value(ShowProgress)>1)
    printf("Time elapsed: ");
printf("%ld:%ld%ld:%ld%ld\n",h,
    m/10, m%10, s/10, s%10);


if (processortime>0)
    {
    h=processortime/3600;
    m=(processortime % 3600)/60;
    s=processortime%60;

    printf("Processor time used %ld:%ld%ld:%ld%ld\n",h,
        m/10, m%10, s/10, s%10);
    }

}        


/*===========================  M I S C ==============================*/


void ModifyCodebookFileHeader(CODEBOOK *CB)
{
char newmethod[MaxGenMethodLength];

ClearGenerationMethod(CB);

sprintf(newmethod, "DA");

AddGenerationMethod(CB, newmethod);
}


/*================  O P E R A T I O N  ==============================*/


void AddPerturbatedDuplicates(TRAININGSET* TS, CODEBOOK* CB, double* Probs)
{
int i, j;
int changed;
int r;

/* Perturbation is done by adding a random number in (-1, 1)
   to each vector component. Obviously not very good for
   binary sets. (0,0,...,0) is not allowed. */

for (i=0; i<BookSize(CB); i++)
    {
    changed = NO;
    while (!changed)
        for (j=0; j<VectorSize(TS); j++)
	    {
	    r = irand(-1,1);
            if ( VectorScalar(CB, i, j) + r >= CB->MinValue &&
		 VectorScalar(CB, i, j) + r <= CB->MaxValue )
		{
		VectorScalar(CB, BookSize(CB) + i, j) =
		    VectorScalar(CB, i, j) + r;
		if (r != 0) changed = YES;
		}
	    }
    Probs[i] = Probs[BookSize(CB) + i] = Probs[i] / 2.0;
    }
BookSize(CB) *= 2;
}


/*--------------------------------------------------------------------*/


double FuzzyMSE(TRAININGSET* TS, CODEBOOK* CB, double** condp)
{
int m, n;
double fMSE = 0.0;

for (m=0; m<BookSize(CB); m++)
    for (n=0; n<BookSize(TS); n++)
	{
	fMSE += condp[m][n] * VectorDist(Vector(TS,n), Vector(CB,m),
			      VectorSize(TS)) * VectorFreq(TS,n);
	}

return fMSE / (VectorSize(TS) * TotalFreq(TS));
}


/*--------------------------------------------------------------------*/


void OptimizeCodebook(TRAININGSET* TS, CODEBOOK* CB, CODEBOOK* newCB,
		      double* Probs, double Beta)
{
int i, m, n;
double **term, *termsum, **condp, *condpsum;
llong mindist;
llong *dists;
int CBsize = BookSize(CB);
double vsum;

/* This should be divided into 2-3 subroutines */

term = (double**) allocate(sizeof(double*) * CBsize);
condp = (double**) allocate(sizeof(double*) * CBsize);
for (i=0; i<CBsize; i++)
    {
    term[i] = (double*) allocate(sizeof(double) * BookSize(TS));
    condp[i] = (double*) allocate(sizeof(double) * BookSize(TS));
    }
termsum = (double*) allocate(sizeof(double) * BookSize(TS));
condpsum = (double*) allocate(sizeof(double) * CBsize);
dists = (llong*) allocate(sizeof(llong) * CBsize);

for (n=0; n<BookSize(TS); n++)
    {
    termsum[n] = 0;
    mindist = MAXLLONG;
    for (m=0; m<CBsize; m++)
	{
        dists[m] = VectorDist(Vector(TS,n), Vector(CB,m), VectorSize(TS));
	if (dists[m] < mindist) mindist = dists[m];
	}
    for (m=0; m<CBsize; m++)
	{
 	term[m][n] = Probs[m] * exp(-dists[m] * Beta);
	termsum[n] += term[m][n];
	}
    }
for (m=0; m<CBsize; m++)
    {
    condpsum[m] = 0;
    for (n=0; n<BookSize(TS); n++)
	{
	condp[m][n] = term[m][n] / termsum[n];
	condpsum[m] += condp[m][n] * VectorFreq(TS, n);
	}
    }
for (m=0; m<CBsize; m++)
    Probs[m] = condpsum[m] / TotalFreq(TS);

CopyCodebookHeader(CB, newCB);
for (m=0; m<CBsize; m++)
    for (i=0; i<VectorSize(TS); i++)
        {
        vsum = 0.0;
	for (n=0; n<BookSize(TS); n++)
	    vsum += VectorScalar(TS, n, i) * VectorFreq(TS, n) * condp[m][n];
        VectorScalar(newCB, m, i) = (VECTORELEMENT) (vsum / condpsum[m] +0.5);
        }
if ( Value(ShowProgress) > 4 )
    printf("fuzzy MSE = %.12f\n", FuzzyMSE(TS, newCB, condp));

deallocate(termsum);
deallocate(condpsum);
for (i=0; i<CBsize; i++)
    {
    deallocate(term[i]);
    deallocate(condp[i]);
    }
deallocate(term);
deallocate(condp);
deallocate(dists);
}


/*--------------------------------------------------------------------*/


YESNO EqualCodebooks(TRAININGSET* TS, CODEBOOK* CB1, CODEBOOK* CB2)
{
int i;

if (BookSize(CB1) != BookSize(CB2)) return NO;
for (i=0; i<BookSize(CB1); i++)
    if ( ! ( EqualVectors( Vector(CB1,i), Vector(CB2,i), VectorSize(TS))))
        return NO;

return YES;
}


/*--------------------------------------------------------------------*/


void IterateCodebook(TRAININGSET* TS, CODEBOOK* CB, double* Probs,
		     double Beta)
{
YESNO goon = YES;
CODEBOOK newCB;
int i, j;

CreateNewCodebook(&newCB, 2*BookSize(CB), TS);

if (BookSize(CB) < Value(CodebookSize))
    AddPerturbatedDuplicates(TS, CB, Probs);
    
while (goon)
    {
    goon = NO;
    OptimizeCodebook(TS, CB, &newCB, Probs, Beta);
    if (!(EqualCodebooks(TS, CB, &newCB))) goon = YES;
    CopyCodebook(&newCB, CB);
    }

for (i=BookSize(CB)-1; i>=0; i--)
    for (j=0; j<i; j++)
	if ( VectorDist(Vector(CB,i), Vector(CB,j), VectorSize(TS)) < THRESHOLD)
	    {
	    if (i < BookSize(CB)-1)
	    CopyVector(Vector(CB,BookSize(CB)-1), Vector(CB,i),
		       VectorSize(TS));
            BookSize(CB)--;
	    break;
	    }
}


/*--------------------------------------------------------------------*/


void IncreaseBeta(double *Beta)
{
(*Beta) *= (1.0 + Value(BetaIncrement) / 10000.0);
}


/*=====================  I N I T I A L I Z E  ==========================*/


void InitializeCodebook(TRAININGSET* TS, CODEBOOK* CB, double** Probs,
		        int* oldCBsize)
{
CreateNewCodebook(CB, 2 * Value(CodebookSize), TS);
*Probs = allocate( sizeof(double) * 2 * Value(CodebookSize) );

CodebookCentroid(TS, Vector(CB, 0));

BookSize(CB) = 1;
*oldCBsize = 0;
    
(*Probs)[0] = 1.0;
}


/*=====================  S H U T D O W N  ============================*/


void ShutDown(TRAININGSET* TS, CODEBOOK* CB, double* Probs)
              
{
FreeCodebook(TS);
FreeCodebook(CB);

deallocate(Probs);
}


/*===========================  M A I N  ==============================*/



int main(int argc, char* argv[])

{
char        TSFileName[80], OutFileName[80];

TRAININGSET TS;
CODEBOOK    CB;
int         oldCBsize;
double*     Probs;
double      Beta;

time_t      StartingTime;

ParameterInfo paraminfo[2] =  { { TSFileName,  FormatNameTS, 0, INFILE  },
                                { OutFileName, FormatNameCB, 0, OUTFILE } };

StartingTime = time(NULL);
ParseParameters( argc, argv, 2, paraminfo);

Beta = 1.0 / Value(InitBetaDivisor);

ReadTrainingSet(TSFileName, &TS); 
/* UnrollTrainingSet(&TS); */

initrandom(Value(RandomSeed));

if (ExistFile(OutFileName) && Value(OverWrite)==NO)
    {
    printf("ERROR: can't write output file\n");
    exit(-1);
    }

InitializeCodebook(&TS, &CB, &Probs, &oldCBsize);

while (Beta < 1.0 / Value(MaxBetaDivisor) )
    {
    IterateCodebook(&TS, &CB, Probs, Beta);
    IncreaseBeta(&Beta);
    if (Value(ShowProgress) > 0)
        printf("Beta = %.6f  CBsize = %d\n", Beta, BookSize(&CB));
    if (Value(ShowProgress) % 2 == 0 && Value(ShowProgress) > 0)
	{
	printf("codebook:\n");
        PrintCodebook(&CB);
	printf("-----------------\n");
	}
    if (BookSize(&CB) > Value(CodebookSize))
	{
	printf("Desired codebook size skipped. Try reducing Beta increment.\n");
        break;
	}
    else if (BookSize(&CB) < oldCBsize)
	{
	printf("Codebook size has diminished. Quitting.\n");
/*        break; */
	}
    else
	{
	if ( Value(ShowProgress) > 2 )
	    printf("MSE = %.6f\n", AverageErrorCB(&TS, &CB, MSE));
	}
    oldCBsize = BookSize(&CB);
    }

/* VectorFreq:t pitisi laittaa kuntoon! */
ModifyCodebookFileHeader(&CB);

WriteCodebook(OutFileName, &CB, Value(OverWrite));
if (Value(ShowProgress) <= 2)
    printf("MSE = %.6f\n", AverageErrorCB(&TS, &CB, MSE));

ShutDown(&TS, &CB, Probs);

if (Value(ShowProgress)>0) OutputTime(StartingTime);
else printf("\n");
checkmemory();
return(0);
}
